import json
import logging
import tarfile
import tempfile
from typing import Any, Callable, Optional, Tuple

from colorama import Fore, Style
from truss.build import create
from truss.local.local_config_handler import (
    LocalConfigHandler as TrussLocalConfigHandler,
)
from truss.truss_handle import TrussHandle

import baseten
from baseten.baseten_deployed_model import BasetenDeployedModel
from baseten.common import api, settings
from baseten.common.core import MODEL_FILENAME, Semver, raises_api_error
from baseten.common.types import TryPatchDraftTrussResponse
from baseten.common.util import base64_encoded_json_str

logger = logging.getLogger(__name__)


def _exists_model(
    model_name: Optional[str], model_id: Optional[str], models_provider: Callable
) -> Optional[Tuple[str, str]]:
    """Checks if an effective model exists for the purpose of deploy or one needs to be created.

    If model name is supplied and a model with that name exists then it's picked up. Otherwise
    working model id in the session is tried.

    Returns id of model if model exists, or None
    """
    models = models_provider()["models"]
    model_id_by_name = {model["name"]: model["id"] for model in models}
    model_name_by_id = {model["id"]: model["name"] for model in models}

    if model_name is not None:
        if model_name in model_id_by_name:
            return model_id_by_name[model_name], model_name
        else:
            logger.warning("Model name not found in deployed models")
            return None

    # No model_name supplied, try to work with working model id
    if model_id is not None:
        if model_id in model_name_by_id:
            return model_id, model_name_by_id[model_id]
        else:
            logger.warning("Working model id not found in deployed models")
            return None

    # No model name or working model
    return None


def build_truss(
    model: Any,
    target_directory: Optional[str] = None,
) -> TrussHandle:
    """
    Builds truss into target directory from in-memory model
    Args:
        model (an in-memory model object): A model object to be deployed (e.g. a keras,
            sklearn, or pytorch model object)
        target_directory (str, optional): The local directory target for the truss.
            Otherwise a temporary directory will be generated
    Returns:
        TrussHandle
    """
    truss = create(model=model, target_directory=target_directory)

    truss_help_messages = f"""
    --------------------------------------------------------------------------------------------
    | {Fore.BLUE} Autogenerating Truss for your model, find more about Truss at https://truss.baseten.co/ |
    --------------------------------------------------------------------------------------------
    You can find your auto generated Truss at {Fore.BLUE} {str(truss.spec.truss_dir.resolve())}
    ---------------------------------------------------------------
    | {Fore.BLUE} Some useful commands to work with your Truss... |
    ---------------------------------------------------------------
    To load your Truss, you can either
    1. Access the Truss handle via the object returned by this function via {Fore.MAGENTA} deployed_model.truss
    2. Load the Truss handle via {Fore.MAGENTA} handle = truss.load('{str(truss.spec.truss_dir.resolve())}')
    To locally run your Truss before deploying, run {Fore.MAGENTA} handle.predict(test_input)
    To add a requirement to your Truss, run {Fore.MAGENTA} handle.add_python_requirement('numpy')
    Visit the docs to learn more about Truss!
    """
    for msg in truss_help_messages.splitlines():
        logger.info(msg.strip())
    return truss


@raises_api_error
def deploy_truss(
    b10_truss: TrussHandle,
    model_name: Optional[str] = None,
    semver_bump: str = Semver.MINOR.value,
    is_trusted: bool = False,
    publish: bool = False,
    external_model_version_id=None,
) -> BasetenDeployedModel:
    """
    Given an existing baseten truss object, deploy it onto the baseten infastructure.

    Args:
        b10_truss: A truss object representing a local baseten_truss directory
        model_name (str, optional): The name of the model to be created, if necessary.
        semver_bump (str, optional): The version bump for this deployment, one of 'MAJOR', 'MINOR', 'PATCH'.
        is_trusted (bool, optional): Whether or not to mark a truss as `trusted` on Baseten.
        publish (bool, optional): Whether or not to mark a truss as `published` on Baseten.
        external_model_version_id (str, optional): If you would rather track model IDs with your own system,
            pass the custom ID here.
    Returns:
        BasetenDeployedModel
    """
    if model_name is None:
        import coolname

        model_name = coolname.generate_slug(2)

    elif not model_name or model_name.isspace():
        raise ValueError("Model Names cannot be empty or spaces")

    # TODO(pankaj) The exists check after generating random name is incorrect.
    # If we happen to generate a random name that's already taken then we will
    # have unexpected behavior.
    model_info = _exists_model(model_name, baseten.working_model_id, api.models)

    if model_info is not None and not publish:
        resp = try_patch_draft_truss(model_name, b10_truss)
        if resp.succeeded or not resp.needs_full_deploy:
            if resp.succeeded:
                logger.info(resp.success_message or f"Model {model_name} patched successfully.")
            else:
                logger.error(
                    f"Failed to patch: `{resp.error_message}`. Model left in original state."
                )

            version_info = resp.model_version_info
            if version_info is None:
                raise RuntimeError("Expecting model version info to be set but found to be None")
            model_id, model_name = model_info
            model_version_id = version_info["id"]
            return BasetenDeployedModel(
                model_version_id=model_version_id, model_name=model_name, truss_handle=b10_truss
            )
        else:
            logger.info(f"Giving up on patching: {resp.error_message}")

    logger.info(f"Serializing {Fore.BLUE}{model_name}{Style.RESET_ALL} truss.")

    # If the truss is not scattered then gather is a no-op and returns path to
    # original truss. So, there's no additional cost to gather for trusses
    # without external packages.
    gathered_truss = TrussHandle(b10_truss.gather())
    temp_file = _compress_truss(b10_truss=gathered_truss)
    logger.info("Making contact with BaseTen 👋 👽")
    s3_key = api.upload_model(temp_file, "tgz", MODEL_FILENAME)
    # String that can be passed through graphql api
    config = base64_encoded_json_str(gathered_truss._spec._config.to_dict())
    if not publish:
        model_id, model_version_id = deploy_draft_truss(
            model_name=model_name,
            s3_key=s3_key,
            config=config,
            is_trusted=is_trusted,
        )
    else:
        if not model_info:
            model_id, model_version_id = create_model_from_s3(
                model_name, s3_key, config, semver_bump, is_trusted, external_model_version_id
            )
            logger.info(f"Successfully registered model {Fore.BLUE}{model_name}{Style.RESET_ALL}.")
        else:
            model_id, model_name = model_info
            model_version_id = create_model_version_from_s3(
                model_id, s3_key, config, semver_bump, is_trusted, external_model_version_id
            )
            logger.info(
                f"Successfully created version {Fore.BLUE}{model_version_id}{Style.RESET_ALL} for {model_name}."
            )

    _generate_model_deploy_logs(model_id)
    return BasetenDeployedModel(
        model_version_id=model_version_id, model_name=model_name, truss_handle=b10_truss
    )


def deploy_draft_truss(
    model_name,
    s3_key,
    config,
    is_trusted=False,
):
    model_version_json = api.deploy_draft_truss(
        model_name=model_name,
        s3_key=s3_key,
        config=config,
        client_version=baseten.__version__,
        is_trusted=is_trusted,
    )

    logger.info(
        f"""
|---------------------------------------------------------------------------------------|
| Your model has been deployed as a draft. Draft models allow you to                    |
| iterate quickly during the deployment process. For more information,                  |
| read https://docs.baseten.co/models/deploying-models/client#stage-2-deploying-a-draft |
|                                                                                       |
| When you are ready to publish your deployed model as a new version,                   |
| pass publish=True to the baseten.deploy() command or publish the version              |
| from the Baseten UI. To skip the draft model step in future deployments,              |
| pass publish=True to the baseten.deploy() command.                                    |
|                                                                                       |
|---------------------------------------------------------------------------------------|

Deployed model info: \n{json.dumps(model_version_json, indent=4)}
"""
    )
    model_id = model_version_json["id"]
    model_version_id = model_version_json["version_id"]
    # Set the newly created model model_version_json
    # be the working model for future commands
    baseten.working_model_id = model_id

    return model_id, model_version_id


def try_patch_draft_truss(
    model_name: str,
    truss_handle: TrussHandle,
) -> TryPatchDraftTrussResponse:
    dev_model_version_info = _get_draft_model_version_info(model_name)
    if dev_model_version_info is None:
        return TryPatchDraftTrussResponse(
            succeeded=False,
            needs_full_deploy=True,
            model_version_info=None,
            error_message=f"No draft version found for model {model_name}.",
        )

    model_deployment_status = dev_model_version_info["current_model_deployment_status"]["status"]
    if model_deployment_status == "BUILDING_MODEL":
        return TryPatchDraftTrussResponse(
            succeeded=False,
            needs_full_deploy=False,
            model_version_info=dev_model_version_info,
            error_message=f"model {model_name} is currently building, please retry after it's finished.",
        )
    if model_deployment_status == "DEPLOYING_MODEL":
        return TryPatchDraftTrussResponse(
            succeeded=False,
            needs_full_deploy=False,
            model_version_info=dev_model_version_info,
            error_message=f"model {model_name} is currently deploying, please retry after it's finished.",
        )

    if model_deployment_status != "MODEL_READY":
        return TryPatchDraftTrussResponse(
            succeeded=False,
            needs_full_deploy=True,
            model_version_info=dev_model_version_info,
            error_message=f"model {model_name} is not ready.",
        )

    if dev_model_version_info is None:
        return TryPatchDraftTrussResponse.needs_full_deploy_response(
            f"no draft model found for model {model_name}."
        )

    # todo: check if model is deploying
    truss_hash = dev_model_version_info["truss_hash"]
    if truss_hash is None:
        return TryPatchDraftTrussResponse.needs_full_deploy_response(
            f"no truss_hash found for model {model_name}."
        )

    truss_signature = dev_model_version_info["truss_signature"]
    if truss_signature is None:
        return TryPatchDraftTrussResponse.needs_full_deploy_response(
            f"no truss_signature found for model {model_name}."
        )

    TrussLocalConfigHandler.add_signature(truss_hash, truss_signature)

    patch_request = truss_handle.calc_patch(truss_hash)
    if patch_request is None:
        return TryPatchDraftTrussResponse.needs_full_deploy_response("unable to calculate patch.")

    if patch_request.prev_hash == patch_request.next_hash:
        return TryPatchDraftTrussResponse.success_response(
            dev_model_version_info, success_message="No changes observed skipping deploy"
        )

    if len(patch_request.patch_ops) == 0:
        # This shouldn't normally happen, hash check should catch this.
        # Fallback to full deploy.
        return TryPatchDraftTrussResponse.needs_full_deploy_response("no patchable changes found.")

    resp = api.patch_draft_truss(
        model_name=model_name,
        client_version=baseten.__version__,
        patch=patch_request,
    )
    if resp["succeeded"] is False:
        needs_full_deploy = True
        if "needs_full_deploy" in resp:
            needs_full_deploy = resp["needs_full_deploy"]
        return TryPatchDraftTrussResponse(
            succeeded=False,
            needs_full_deploy=needs_full_deploy,
            model_version_info=dev_model_version_info,
            error_message=resp["error"],
        )

    return TryPatchDraftTrussResponse.success_response(
        dev_model_version_info,
        success_message=f"Successfully patched draft version of model `{model_name}`.",
    )


def _get_draft_model_version_info(model_name) -> Optional[dict]:
    model = api.get_model(model_name)
    versions = model["model_version"]["oracle"]["versions"]
    for version in versions:
        if version["is_draft"] is True:
            return version
    return None


def create_model_from_s3(
    model_name,
    s3_key,
    config,
    semver_bump,
    is_trusted=False,
    external_model_version_id=None,
):
    model_version_json = api.create_model_from_truss(
        model_name=model_name,
        s3_key=s3_key,
        config=config,
        semver_bump=semver_bump,
        client_version=baseten.__version__,
        is_trusted=is_trusted,
        external_model_version_id=external_model_version_id,
    )

    logger.info(f"Created model:\n{json.dumps(model_version_json, indent=4)}")
    model_id = model_version_json["id"]
    model_version_id = model_version_json["version_id"]
    # Set the newly created model model_version_json
    # be the working model for future commands
    baseten.working_model_id = model_id

    return model_id, model_version_id


def create_model_version_from_s3(
    model_id,
    s3_key,
    config,
    semver_bump,
    is_trusted=False,
    external_model_version_id=None,
):
    model_version_json = api.create_model_version_from_truss(
        model_id=model_id,
        s3_key=s3_key,
        config=config,
        semver_bump=semver_bump,
        client_version=baseten.__version__,
        is_trusted=is_trusted,
        external_model_version_id=external_model_version_id,
    )
    model_version_id = model_version_json["id"]
    return model_version_id


def _generate_model_deploy_logs(model_id):
    logger.info(f"{Fore.BLUE} Deploying model version.")
    model_version_web_url = f"{settings.get_server_url()}/models/{model_id}"
    logger.info("🏁 The model is being built and deployed right now 🏁")
    visit_message = (
        f"|  Visit {Fore.BLUE}{model_version_web_url}{Style.RESET_ALL} for deployment status  |"
    )
    visit_message_len = len(visit_message) - len(Fore.BLUE) - len(Style.RESET_ALL)
    logger.info("".join(["-" for _ in range(visit_message_len)]))
    logger.info(visit_message)
    logger.info("".join(["-" for _ in range(visit_message_len)]))


def _create_temporary_file(b10_truss, delete=True):
    temp_file = tempfile.NamedTemporaryFile(suffix=".tgz", delete=delete)
    with tarfile.open(temp_file.name, "w:gz") as tar:
        tar.add(b10_truss._spec.truss_dir, arcname=".")
    return temp_file


def _compress_truss(b10_truss):
    try:
        temp_file = _create_temporary_file(b10_truss)
    except PermissionError:
        # Windows bug with Tempfile causes PermissionErrors
        temp_file = _create_temporary_file(b10_truss, delete=False)
    temp_file.file.seek(0)
    return temp_file


def build_and_deploy_truss(
    model: Any,
    model_name: Optional[str] = None,
    version_bump=Semver.MINOR.value,
    is_trusted=False,
    publish=False,
    external_model_version_id=None,
) -> BasetenDeployedModel:
    if isinstance(model, TrussHandle):
        return deploy_truss(
            model, model_name, version_bump, is_trusted, publish, external_model_version_id
        )
    truss = build_truss(model)
    return deploy_truss(
        truss, model_name, version_bump, is_trusted, publish, external_model_version_id
    )


def pull_model(model_version_id: str, directory: str = "."):
    deployed_model = BasetenDeployedModel(model_version_id=model_version_id)
    return deployed_model.pull(directory)



import base64
from io import BytesIO
from pickle import UnpicklingError
from typing import Any, Optional, Tuple

from PIL import Image

from baseten.baseten_deployed_model import BasetenDeployedModel
from baseten.common import api, settings
from baseten.common.files import upload_pil_to_s3
from baseten.models.util import (
    from_object_to_str,
    from_str_to_object,
    get_or_create_pretrained_model,
    requests_error_handling,
)


class StableDiffusionPipeline:
    """
    Support for Stable Diffusion for text-to-image generation.


    **Examples:**

    Invoking default Stable Diffusion

    ```python
    from baseten.models import StableDiffusionPipeline
    model = StableDiffusionPipeline()
    images, is_nsfw, url = model("A dog is running in the grass")
    ```

    **Using a scheduler**

    Schedulers can significantly influence how your generated image turns out and the quality of the image.
    Blueprint provides sensible defaults for a set of common schedulers but you can also define your own scheduler.

    ```python
    from baseten.models import StableDiffusionPipeline
    from diffusers import LMSDiscreteScheduler

    lms = LMSDiscreteScheduler()
    model = StableDiffusionPipeline(scheduler=lms)

    images, is_nsfw, url = model("A dog is running in the grass")
    ```

    **Using a VAE**

    A variational auto encoder (VAE) is used within Stable Diffusion to help construct images from prompts.
    It can have a significant effect on the output of the model. Blueprint supports utilizing 3 VAEs as shown below.
    You may want to experiment with a new VAE if you want to improve the generation of hands and faces.

    ```python
    from baseten.models import StableDiffusionPipeline

    model = StableDiffusionPipeline(vae="stabilityai/sd-vae-ft-mse")

    images, is_nsfw, url = model("A dog is running in the grass")
    ```

    You can combine a scheduler and VAE for even more control.

    ```python
    from baseten.models import StableDiffusionPipeline

    model = StableDiffusionPipeline(
        "runwayml/stable-diffusion-v1-5",
        vae="stabilityai/sd-vae-ft-mse",
        scheduler="pndm"
    )

    images, is_nsfw, url = model("A dog is running in the grass", seed=2)
    ```
    """

    def __init__(
        self,
        hf_pretrained_model_path: str = "runwayml/stable-diffusion-v1-5",
        vae: str = "default",
        scheduler: str = "dpmsolver",
        model_id: Optional[str] = None,
    ):
        """
        Args:
            hf_pretrained_model_path: Name of the Hugging Face pretrained model to use for text-to-image generation.
                By default, we will use the v1.5 version provided by Huggingface.
            vae:
                Name of the pretrained VAE model to use for image reconstruction.
                By default, we will use the "default" VAE. Users can also specify a custom VAE model either

                - `stabilityai/sd-vae-ft-mse`
                - `stabilityai/sd-vae-ft-ema`

            scheduler (str | SchedulerMixin):
                Name of the scheduler to use for the diffusion process. By default, we will use
                the "dpmsolver" scheduler. Users can also specify a custom scheduler that is either
                one of

                - `dpmsolver`
                - `ddim`
                - `euler`
                - `lms`
                - `pndm`
                - `eulera`

                A user can also specify a custom scheduler by passing in a scheduler object that
                is a subclass of the diffusers.scheduler Mixin.

            model_id (Optiona;[str]):
                ID of the finetuned model to use for text-to-image generation. By default,
                we will use the generic Stable Diffusion model.
        """

        self._supported_alterations = {
            "model": ["runwayml/stable-diffusion-v1-5"],
            "vae": ["stabilityai/sd-vae-ft-mse", "stabilityai/sd-vae-ft-ema", "default"],
            "scheduler": ["ddim", "euler", "lms", "pndm", "eulera", "dpmsolver"],
        }
        self._hf_pretrained_model_path = hf_pretrained_model_path
        self._id = model_id
        self.vae = vae
        self.scheduler = scheduler
        self.safety_checker = None
        self.feature_extractor = None
        self._is_user_model = self._id is not None
        self._model = self._set_user_model()
        self._is_model_components_supported()

    def _set_user_model(self) -> BasetenDeployedModel:
        """Creates internal BasetenDeployedModel object that points to users
        deployed Stable Diffusion model. If the user does not have a deployed
        model, we will create one for them.
        """
        if self._is_user_model:
            return BasetenDeployedModel(model_id=self._id)
        model_version = get_or_create_pretrained_model("Stable Diffusion")
        return BasetenDeployedModel(
            model_id=model_version["model_id"],
            truss_spec_version=model_version["truss_spec_version"],
        )

    def _is_supported_scheduler(self, scheduler) -> None:
        """Checks if the scheduler is supported by the BasetenDeployedModel"""
        if self._is_user_model and scheduler != "dpmsolver":
            raise ValueError("Cannot specify scheduler for a finetuned model")
        supported_schedulers = self._supported_alterations["scheduler"]
        if scheduler not in supported_schedulers:
            # Check if scheduler is from diffusers.scheduler
            if "diffusers.scheduler" not in str(type(self.scheduler)):
                raise ValueError(
                    f"""
                    Scheduler must be one of {' or '.join(self._supported_alterations['scheduler'])}
                    or a Diffusers scheduler
                """
                )

    def _is_supported_vae(self, vae) -> None:
        """Checks if the vae is supported by the BasetenDeployedModel"""
        if self._is_user_model and vae != "default":
            raise ValueError("Cannot specify VAE for a finetuned model")
        if vae not in self._supported_alterations["vae"]:
            raise ValueError(
                f"""
                VAE must be one of {' or '.join(self._supported_alterations['vae'])} as str
            """
            )

    def _is_supported_model(self, model) -> None:
        """Checks if the model version is supported by the BasetenDeployedModel"""
        if not self._is_user_model:
            if model not in self._supported_alterations["model"]:
                raise ValueError(
                    f"""
                    Pretrained model path must be one of {' or '.join(self._supported_alterations['model'])}
                """
                )

    def _is_model_components_supported(self) -> None:
        """Checks if model components are supported by the BasetenDeployedModel
        before request to the model is made.
        """
        self._is_supported_model(self._hf_pretrained_model_path)
        self._is_supported_vae(self.vae)
        self._is_supported_scheduler(self.scheduler)

    def __setattr__(self, __name: str, __value: Any) -> None:
        """Prevents users from setting unsupported model components"""
        if __name in ["unet", "tokenizer", "text_encoder"]:
            raise ValueError(f"{__name} not settable for Blueprint StableDiffusionPipeline models")
        super().__setattr__(__name, __value)

    def _handle_generic_model_response(self, response: dict) -> Tuple:
        """Handles the response from a generic Stable Diffusion model and returns
        the image, safe boolean and s3 link.

        Args:
            response (dict): Response from the model

        Returns:
            Tuple: Tuple containing the image, safe boolean and s3 link
        """
        try:
            output = from_str_to_object(response["data"][0])
        except UnpicklingError:
            raise ValueError(
                "Error unpickling response. This may be due to an older \
                Python version. If so, please `pip install pickle5` and try again."
            )
        except ValueError:
            raise ValueError("Corruputed network bytes")

        # The server returns a Tuple of (PIL, bool)
        if output[1] is None:
            pil_objects = output[0]
            s3_urls = [upload_pil_to_s3(pil_object) for pil_object in pil_objects]

            return (output[0], False, s3_urls)
        return output

    def _handle_user_model_response(self, response: dict) -> Tuple:
        """Handles the response from a user's finetuned Stable Diffusion model and returns
        the image, safe boolean and s3 link.

        Args:
            response (dict): Response from the model

        Returns:
            Tuple: Tuple containing the image, safe boolean and s3 link
        """
        model_response = response["data"][0]
        # decode base64 string into image
        # TODO(abu): add support for multiple images
        image = Image.open(BytesIO(base64.b64decode(model_response)))
        s3_url = upload_pil_to_s3(image)
        return (image, False, [s3_url])

    def __call__(
        self,
        prompt: str,
        height: int = 512,
        width: int = 512,
        guidance_scale: float = 7.5,
        num_images_per_prompt: int = 1,
        eta: float = 0.0,
        seed: int = -1,
        num_inference_steps: int = -1,
    ) -> Tuple:
        """Generate image from a prompt. Supports all parameters of the underlying model
        from diffusers library.

        Args:
            prompt: Text prompt to generate image from
            height: Height of the generated image. Defaults to 512.
            width: Width of the generated image. Defaults to 512.
            guidance_scale (float, optional): Guidance scale for the diffusion process.
                Defaults to 7.5.
            num_images_per_prompt (int, optional): Number of images to generate per prompt.
                Defaults to 1.
            eta (float, optional): Eta value for the diffusion process. Defaults to 0.0.
            seed (int, optional): Seed for the diffusion process. Defaults to -1, which will
                use a random seed.
            num_inference_steps (int, optional): Number of inference steps for the
                diffusion process. Defaults to None.

        Returns:
            generated_images:
                A tuple containing three values:

                - `images`: A list of `PIL.Image` images
                - `is_nsfw`: `True` if adult content is detected in returned images, else `false`
                - `urls`: A list of URLs linking the generated images on S3
        """
        self._is_model_components_supported()

        # If default scheduler, DPMSolver, is being used, set num_inference_steps to 30
        # instead of 50 as it requires less time steps for high quality generation
        if num_inference_steps < 0:
            num_inference_steps = 30 if self.scheduler == "DPMSolver" else 50

        request_body = {
            "prompt": prompt,
            "height": height,
            "width": width,
            "guidance_scale": guidance_scale,
            "num_images_per_prompt": num_images_per_prompt,
            "num_inference_steps": num_inference_steps,
            "eta": eta,
            "seed": seed,
            "scheduler": self.scheduler
            if isinstance(self.scheduler, str)
            else from_object_to_str(self.scheduler),
            "vae": self.vae,
            "b64_response": False,
        }
        with requests_error_handling():
            server_response = self._model.predict(request_body)

        if server_response["status"] == "error":
            raise ValueError(server_response["message"])

        if self._is_user_model:
            return self._handle_user_model_response(server_response)

        return self._handle_generic_model_response(server_response)

    @property
    def blueprint_url(self) -> str:
        """
        If this `StableDiffusionPipeline` is your fine-tuned model, use this property to
        get a link to the Blueprint page with your model information. Otherwise, it gives
        you the Stable Diffusion model page, same as `StableDiffusionPipeline.url()`.

        **Example:**

        ```python
        from baseten.models import StableDiffusionPipeline

        model = StableDiffusionPipeline(model_id="MODEL_ID")
        model.blueprint_url
        ```
        """
        if self._is_user_model:
            return self._model.blueprint_url
        return StableDiffusionPipeline.url()

    @staticmethod
    def url():
        """
        Use this static method to get a URL to the Stable Diffusion model page
        in your Blueprint project, which contains information about the model.

        **Example:**

        ```python
        from baseten.models import StableDiffusionPipeline

        StableDiffusionPipeline.url()
        ```
        """
        try:
            blueprint_project_id = api.get_blueprint_projects()[0]["id"]
            return (
                f"{settings.get_server_url()}/blueprint/projects/"
                f"{blueprint_project_id}?est=community-model--stable-diffusion"
            )
        except IndexError:
            raise ValueError("User does not have any blueprint projects.")



import uuid
from contextlib import contextmanager
from typing import Any, Dict, Optional

import cloudpickle
import requests

from baseten.common import api
from baseten.common.core import AuthorizationError

MODEL_ZOO_CONFIG = {
    "Stable Diffusion": "stable_diffusion",
    "Whisper": "whisper",
    "Flan-T5 XL": "flan_t5",
}


def from_str_to_object(inp: str) -> Any:
    """Converts a hex string that represents a Cloudpickled Python object to a Python object.

    Args:
        inp (str): Hex string that represents a Cloudpickled Python object

    Returns:
        Any: Python object
    """
    byte_stream = bytes.fromhex(inp)
    return cloudpickle.loads(byte_stream)


def from_object_to_str(inp) -> str:
    """Converts a Python object to a hex string that represents a Cloudpickled Python object.

    Args:
        inp (Any): Python object

    Returns:
        str: Hex string that represents a Cloudpickled Python object
    """
    return cloudpickle.dumps(inp).hex()


def get_pretrained_model(model_name: str) -> Optional[Dict]:
    """Checks a users pretrained models models to see if they have a specific model.
    Models are also checked to be in healthy condition. If they are not, we do not
    return them.

    Healthy models are 1 of 2 cases:
    1. The model has a primary version and that primary version is healthy. If a primary
    version is not healthy, we do not return the model. If a primary version exists,
    Django will route all requests to it by default so we need to make sure it is healthy.
    2. The model does not have a primary version, but it has a healthy version.

    Args:
        model_name (str): Name of the model to check for

    Returns:
        Optional[Dict]: Model version if model exists, else None
    """
    try:
        pretrained_model_version_api_response = api.pretrained_model_version(
            pretrained_model_name=MODEL_ZOO_CONFIG[model_name]
        )
    except AuthorizationError:
        raise AuthorizationError("You must first run the `baseten signup` command")
    # See if the model exists in users models

    pretrained_model_version = pretrained_model_version_api_response.get("pretrained_model_version")

    if pretrained_model_version:
        status = pretrained_model_version["current_deployment_status"]
        if status in ["MODEL_READY", "DEPLOYING_MODEL"]:
            return pretrained_model_version
        # If a model is set as primary and is not ready, Django will route requests to the
        # failed model so we need to create a new model version so we return None.
        return None

    return None


def create_pretrained_model(model_name: str) -> dict:
    """Creates a model from the model zoo and returns the model id.

    Args:
        model_name (str): Name of the model to create

    Returns:
        dict: model
    """
    return api.create_pretrained_model(MODEL_ZOO_CONFIG[model_name], model_name)


def get_or_create_pretrained_model(model_name: str) -> Dict:
    """Checks if a model exists in the users models. If it does not, it creates it.

    Args:
        model_name (str): Name of the model to check for

    Returns:
        str: Model id
    """
    if model_name not in MODEL_ZOO_CONFIG:
        raise ValueError("Model {model_name} not supported")
    with requests_error_handling():
        model_version = (
            get_pretrained_model(model_name)
            or create_pretrained_model(model_name)["primaryVersion"]
        )

    return model_version


def upload_file_to_s3(file_path: str) -> str:
    """Uploads a file to S3 and returns the S3 url.

    Args:
        file_path (str): Path to file to upload

    Returns:
        str: The URL to get that file from
    """
    file_extension = file_path.split(".")[-1]
    temp_data_name = f"{str(uuid.uuid4())}.{file_extension}"

    signed_s3_upload_post = api.signed_s3_upload_and_download_post(temp_data_name)
    file_data = {"file": (temp_data_name, open(file_path, "rb"))}

    form_fields = signed_s3_upload_post["form_fields"]
    form_fields["AWSAccessKeyId"] = form_fields.pop("aws_access_key_id")

    with requests_error_handling():
        requests.post(signed_s3_upload_post["url"], data=form_fields, files=file_data)
    return signed_s3_upload_post["get_url"]


@contextmanager
def requests_error_handling():
    """Context manager to handle Requests errors"""
    try:
        yield
    except requests.exceptions.HTTPError as errh:
        print("ERROR: We encountered an HTTP error: ", errh)
    except requests.exceptions.ConnectionError as errc:
        print("ERROR: We are having trouble connecting:", errc)
    except requests.exceptions.Timeout as errt:
        print("ERROR: The request to the model has timed out:", errt)
    except requests.exceptions.RequestException as err:
        print("ERROR: Something went wrong, please try again:", err)



from pathlib import Path

import click

from baseten.common.files import DatasetTrainingType, upload_dataset


@click.group(name="dataset")
def dataset_cli():
    """Manage datasets stored on Blueprint"""
    pass


@dataset_cli.command()
@click.argument("data_dir", type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option("--name", "-n", required=True, type=str, help="Name give to Dataset on Blueprint")
@click.option(
    "--training-type",
    "-t",
    required=True,
    type=click.Choice([t.value for t in DatasetTrainingType], case_sensitive=False),
    help="Types of FinetuningRun that can use this Dataset",
)
def upload(data_dir: Path, name: str, training_type: str):
    """Upload a new dataset to Baseten from DATA_DIR

    DATA_DIR is the path to directory with relevant files.
    """
    dataset_id = upload_dataset(name, Path(data_dir), DatasetTrainingType[training_type.upper()])
    click.echo()
    click.echo(f"Dataset ID:\n{dataset_id}")



import logging
import time
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, List, Optional

import pytz
from halo import Halo

from baseten.common.api import get_training_logs

# Needed to avoid circular dependencies
if TYPE_CHECKING:
    from baseten.training.finetuning import FinetuningRun


# These two parameters are used for the backend query.
# Required for correctness and performance.
# Defaults assume that no more than 1000 lines are emitted in 5 minutes of training.
# TODO: make settable by client for more verbose jobs.
_TRAINING_LOG_QUERY_TIMESTEP = timedelta(minutes=5)
_TRAINING_LOG_QUERY_LIMIT = 1000

# Constants for streaming logs
_PENDING_STATE_DELAY_SECONDS = 1.0
_RUNNING_STATE_DELAY_SECONDS = 0.2


logger = logging.getLogger(__name__)


@dataclass
class LogLine:
    """
    Wrapper for log lines.
    """

    ts: int
    msg: str
    # level is optional since it's currently not handle on the server side
    level: Optional[str]

    def console_stream(self):
        # TODO: update based on level
        logger.info(self.msg)


def _typed_and_ordered_training_logs(
    training_run_id: str,
    start: Optional[str] = None,
    end: Optional[str] = None,
    limit: Optional[int] = None,
) -> List[LogLine]:
    inner_logs = get_training_logs(training_run_id, start, end, limit=limit)
    return sorted([LogLine(**d) for d in inner_logs], key=lambda logline: int(logline.ts))


def _dt_millis_str(dt: datetime) -> str:
    return str(int(dt.timestamp() * 1000))


def _dt_from_nanos_str(timestamp: int) -> datetime:
    return datetime.fromtimestamp(float(timestamp) / 1e9).replace(tzinfo=pytz.UTC)


class TrainingLogsConsumer:
    def __init__(self, run: "FinetuningRun") -> None:
        self._run = run
        self._entity_id: str = run.id
        self._refresh()

    def _refresh(self):
        self._run.refresh()
        self._start_time: Optional[datetime] = (
            datetime.fromisoformat(self._run.started).replace(tzinfo=pytz.UTC)
            if self._run.started
            else None
        )
        self._stop_time: Optional[datetime] = (
            datetime.fromisoformat(self._run.stopped).replace(tzinfo=pytz.UTC)
            if self._run.stopped
            else None
        )

    def _get_lastest_time(self) -> datetime:
        return self._stop_time if self._stop_time else datetime.now()

    def _consume_logs_period(
        self, start_time: datetime, end_time: datetime, limit: Optional[int] = None
    ) -> List[LogLine]:
        if start_time is None:
            return []
        end_time = end_time.replace(tzinfo=pytz.UTC)
        query_start = start_time
        query_end = min(start_time + _TRAINING_LOG_QUERY_TIMESTEP, end_time)
        if start_time > end_time:
            return []
        logs = []
        while query_end <= end_time:
            logs.extend(
                _typed_and_ordered_training_logs(
                    self._entity_id,
                    start=_dt_millis_str(query_start),
                    end=_dt_millis_str(query_end),
                    limit=limit or _TRAINING_LOG_QUERY_LIMIT,
                )
            )

            # Range is inclusive so we have to increment up slightly
            # NOTE: this is slightly error prone due to the discrepency of
            #       loki working in nanos and graphql taking millis
            query_start = query_end + timedelta(milliseconds=1)
            query_end = query_start + _TRAINING_LOG_QUERY_TIMESTEP
        return logs

    def _get_next_start_time(self, last_logs: List[LogLine], last_start_time: datetime) -> datetime:
        if len(last_logs) > 0:
            # Range is inclusive so we have to increment up slightly
            # NOTE: this is slightly error prone due to the discrepency of
            #       loki working in nanos and graphql taking millis
            return _dt_from_nanos_str(last_logs[-1].ts) + timedelta(milliseconds=1)
        return last_start_time

    def stream_logs(self):
        if self._run.is_pending:
            spinner = Halo(text="Waiting for finetuning job to start", spinner="dots")
            spinner.start()

            while self._run.is_pending or self._start_time is None:
                time.sleep(_PENDING_STATE_DELAY_SECONDS)
                self._refresh()

            spinner.succeed("Finetuning job successully started running.")

        Halo().info("Starting to stream logs")

        retrieved_logs = self._consume_logs_period(self._start_time, self._get_lastest_time())
        period_start_time = self._get_next_start_time(retrieved_logs, self._start_time)
        for line in retrieved_logs:
            line.console_stream()

        while self._run.is_running:
            retrieved_logs = self._consume_logs_period(period_start_time, self._get_lastest_time())
            period_start_time = self._get_next_start_time(retrieved_logs, period_start_time)
            for line in retrieved_logs:
                line.console_stream()

            time.sleep(_RUNNING_STATE_DELAY_SECONDS)
            self._refresh()

        for line in self._consume_logs_period(period_start_time, self._get_lastest_time()):
            line.console_stream()

        if self._run.is_succeeded:
            Halo().succeed("Finetuning job completed successfully.")
        elif self._run.is_failed:
            Halo().fail("Finetuning job failed.")
        elif self._run.is_cancelled:
            Halo().warn("Finetuning job was cancelled.")



import functools
import json
import logging
import os
import tarfile
import tempfile
from typing import IO, Dict, List, Optional, Tuple, Union

import requests
from colorama import Fore, Style
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tqdm import tqdm

from baseten.common import settings
from baseten.common.core import ARTIFACT_FILENAME, ApiError, AuthorizationError
from baseten.common.util import base64_encoded_json_str

logger = logging.getLogger(__name__)


class AuthToken:
    value: str

    def __init__(self, value: str):
        self.value = value

    def headers(self):
        raise NotImplementedError


class ApiKey(AuthToken):
    def headers(self):
        return {"Authorization": f"Api-Key {self.value}"}


class JWT(AuthToken):
    def headers(self):
        return {"Authorization": self.value}


def with_api_key(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        config = settings.read_config()

        if config.has_option("api", "api_key"):
            api_key = config.get("api", "api_key")
        elif settings.get_jwt_auth_token():
            raise AuthorizationError("This operation is not supported in the Web IDE")
        else:
            raise AuthorizationError("You must first run the `baseten login` cli command.")

        result = func(ApiKey(api_key), *args, **kwargs)
        return result

    return wrapper


def with_api_key_or_jwt(func):
    """
    Allows an API method to be called internally (with JWT) or externally
    (with API Key).
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        config = settings.read_config()

        if config.has_option("api", "api_key"):
            result = func(ApiKey(config.get("api", "api_key")), *args, **kwargs)
            return result
        elif settings.get_jwt_auth_token():
            result = func(JWT(settings.get_jwt_auth_token()), *args, **kwargs)
            return result
        else:
            raise AuthorizationError("You must first run the `baseten login` cli command.")

    return wrapper


@with_api_key_or_jwt
def models(auth_token: AuthToken):
    query_string = """
    {
      models {
        id,
        name
        versions{
            id,
            semver,
            current_deployment_status,
            is_primary,
        }
      }
    }
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]


@with_api_key
def get_model(api_key, model_name):
    query_string = f"""
    {{
      model_version(name: "{model_name}") {{
        oracle{{
            name
            versions{{
                id
                semver
                truss_hash
                truss_signature
                is_draft
                current_model_deployment_status {{
                    status
                }}
            }}
        }}
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]


@with_api_key
def get_pretrained_model(api_key, pretrained_model_name):
    query_string = f"""
    {{
      pretrained_model(name: "{pretrained_model_name}") {{
        pretty_name
        name
        s3_key
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]


@with_api_key
def pretrained_models(api_key):
    query_string = """
    {
      pretrained_models {
        pretty_name
        name
        s3_key
      }
    }
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]


@with_api_key_or_jwt
def pretrained_model_version(auth_token: AuthToken, pretrained_model_name: str):
    query_string = f"""
    {{
      pretrained_model_version(pretrained_model_name: "{pretrained_model_name}"){{
        model_id
        current_deployment_status
        truss_spec_version
      }}
    }}
    """
    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]


@with_api_key
def models_summary(api_key):
    query_string = """
    {
      models {
        id,
        name,
        description,
        versions {
          id
          created,
          configuration
        }
      }
    }
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["models"]


@with_api_key
def model_version_external_id_get_version(api_key, external_model_version_id):
    query_string = f"""
    {{
      model_version(external_model_version_id: "{external_model_version_id}") {{
        id
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model_version"]["id"]


@with_api_key_or_jwt
def model_version_truss_spec_version(auth_token: AuthToken, model_version_id: str):
    query_string = f"""
    {{
      model_version(id: "{model_version_id}") {{
        truss_spec_version
      }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["model_version"]["truss_spec_version"]


@with_api_key_or_jwt
def model_truss_spec_version(auth_token: AuthToken, model_id: str):
    query_string = f"""
    {{
      model(id: "{model_id}") {{
        primary_version {{
            truss_spec_version
        }}
      }}
    }}
    """
    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["model"]["primary_version"]["truss_spec_version"]


@with_api_key
def create_model_from_truss(
    api_key,
    model_name,
    s3_key,
    config,
    semver_bump,
    client_version,
    is_trusted=False,
    external_model_version_id=None,
):
    query_string = f"""
    mutation {{
      create_model_from_truss(name: "{model_name}",
                   s3_key: "{s3_key}",
                   config: "{config}",
                   semver_bump: "{semver_bump}",
                   client_version: "{client_version}",
                   is_trusted: {'true' if is_trusted else 'false'}
                   external_model_version_id: "{external_model_version_id if external_model_version_id else ''}"
) {{
        id,
        name,
        version_id
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["create_model_from_truss"]


@with_api_key
def create_model_version_from_truss(
    api_key,
    model_id,
    s3_key,
    config,
    semver_bump,
    client_version,
    is_trusted=False,
    external_model_version_id=None,
):

    query_string = f"""
    mutation {{
      create_model_version_from_truss(
                   model_id: "{model_id}"
                   s3_key: "{s3_key}",
                   config: "{config}",
                   semver_bump: "{semver_bump}",
                   client_version: "{client_version}",
                   is_trusted: {'true' if is_trusted else 'false'}
                   external_model_version_id: "{external_model_version_id if external_model_version_id else ''}"

      ) {{
        id,
        name,
        version_id
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["create_model_version_from_truss"]


@with_api_key
def deploy_draft_truss(
    api_key,
    model_name,
    s3_key,
    config,
    client_version,
    is_trusted=False,
):
    query_string = f"""
    mutation {{
      deploy_draft_truss(name: "{model_name}",
                   s3_key: "{s3_key}",
                   config: "{config}",
                   client_version: "{client_version}",
                   is_trusted: {'true' if is_trusted else 'false'},
) {{
        id,
        name,
        version_id
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["deploy_draft_truss"]


@with_api_key
def patch_draft_truss(
    api_key,
    model_name,
    patch,
    client_version,
):
    patch = base64_encoded_json_str(patch.to_dict())
    query_string = f"""
    mutation {{
      patch_draft_truss(name: "{model_name}",
                   client_version: "{client_version}",
                   patch: "{patch}",
) {{
        id,
        name,
        version_id
        succeeded
        needs_full_deploy
        error
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["patch_draft_truss"]


@with_api_key
def create_model_configuration(
    api_key: ApiKey,
    configuration: dict,
    model_id: Optional[str] = None,
    model_version_id: Optional[str] = None,
):
    # encode for None/null
    encoded_configuration = base64_encoded_json_str(configuration)

    model_identifier_field = None
    if model_id:
        model_identifier_field = f'model_id: "{model_id}"'
    elif model_version_id:
        model_identifier_field = f'model_version_id: "{model_version_id}"'
    else:
        raise ValueError("Either model_version_id or model_id must be specified.")

    query_string = f"""
    mutation {{
        create_model_configuration(
            encoded_configuration: "{encoded_configuration}",
            {model_identifier_field}
        )
        {{
            configuration_id: id,
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["create_model_configuration"]


@with_api_key
def model_version_configuration(api_key, model_version_id):
    query_string = f"""
    {{
      model_version(id: "{model_version_id}") {{
        configuration
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model_version"]["configuration"]


@with_api_key
def model_configuration(api_key, model_id):
    query_string = f"""
    {{
      model(id: "{model_id}") {{
        configuration
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model"]["configuration"]


@with_api_key
def create_model_metrics(
    api_key: ApiKey,
    metric_data: dict,
    model_id: Optional[str] = None,
    model_version_id: Optional[str] = None,
):
    # encode for None/null
    encoded_metric_data = base64_encoded_json_str(metric_data)

    model_identifier_field = None
    if model_id:
        model_identifier_field = f'model_id: "{model_id}"'
    elif model_version_id:
        model_identifier_field = f'model_version_id: "{model_version_id}"'
    else:
        raise ValueError("Either model_version_id or model_id must be specified.")

    query_string = f"""
    mutation {{
        create_model_metrics(
            encoded_metric_data: "{encoded_metric_data}",
            {model_identifier_field}
        )
        {{
            metrics_id: id,
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["create_model_metrics"]


@with_api_key
def model_version_metrics(api_key, model_version_id):
    query_string = f"""
    {{
      model_version(id: "{model_version_id}") {{
        metrics
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model_version"]["metrics"]


@with_api_key
def model_metrics(api_key, model_id):
    query_string = f"""
    {{
      model(id: "{model_id}") {{
        metrics
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model"]["metrics"]


@with_api_key
def signed_s3_upload_post(api_key, model_file_name):
    query_string = f"""
    {{
      signed_s3_upload_url(model_file_name: "{model_file_name}") {{
        url,
        form_fields {{
          key,
          aws_access_key_id,
          policy,
          signature,
        }}
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["signed_s3_upload_url"]


@with_api_key
def signed_s3_upload_and_download_post(api_key, file_name):
    query_string = f"""
    {{
      signed_s3_upload_url(model_file_name: "{file_name}"){{
        url,
        get_url,
        form_fields {{
          key,
          aws_access_key_id,
          policy,
          signature,
        }}
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["signed_s3_upload_url"]


@with_api_key
def get_model_s3_download_url(api_key, model_version_id):
    query_string = f"""
    {{
      model_s3_download_url(model_version_id: "{model_version_id}") {{
        url,
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model_s3_download_url"]["url"]


@with_api_key
def artifact_signed_s3_upload_post(api_key, file_name, file_size):
    query_string = f"""
    {{
      artifact_signed_s3_upload_url(file_name: "{file_name}", file_size: {file_size}) {{
        url,
        form_fields {{
          key,
          aws_access_key_id,
          policy,
          signature,
        }}
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["artifact_signed_s3_upload_url"]


@with_api_key_or_jwt
def user_files_signed_s3_upload_post(auth_token: AuthToken, file_name: str):
    query_string = f"""
    {{
      user_files_signed_s3_upload_url(file_name: "{file_name}") {{
        url,
        get_url,
        form_fields {{
          key,
          aws_access_key_id,
          policy,
          signature,
        }}
      }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)

    return resp["data"]["user_files_signed_s3_upload_url"]


@with_api_key
def register_data_for_model(api_key, s3_key, model_version_id, data_name):
    query_string = f"""
    mutation {{
        create_sample_data_file(model_version_id: "{model_version_id}",
                                name: "{data_name}",
                                s3_key: "{s3_key}") {{
          id
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["create_sample_data_file"]


@with_api_key
def create_artifact(
    api_key: ApiKey,
    s3_key: str,
    name: str,
    description: Optional[str] = None,
    model_id: Optional[str] = None,
    model_version_id: Optional[str] = None,
):
    # encode for None/null
    description = base64_encoded_json_str(description)
    model_id = base64_encoded_json_str(model_id)
    model_version_id = base64_encoded_json_str(model_version_id)

    query_string = f"""
    mutation {{
        create_artifact(
            name: "{name}",
            s3_key: "{s3_key}",
            encoded_description: "{description}",
            encoded_model_id: "{model_id}",
            encoded_model_version_id: "{model_version_id}"
        )
        {{
            artifact_id: id,
            name,
            description
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["create_artifact"]


@with_api_key
def create_artifact_link(
    api_key: ApiKey,
    artifact_id: str,
    model_id: Optional[str] = None,
    model_version_id: Optional[str] = None,
):
    # encode for None/null
    model_id = base64_encoded_json_str(model_id)
    model_version_id = base64_encoded_json_str(model_version_id)

    query_string = f"""
    mutation {{
        link_artifact(
            id: "{artifact_id}",
            encoded_model_id: "{model_id}",
            encoded_model_version_id: "{model_version_id}"
        )
        {{
            id
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["link_artifact"]


@with_api_key
def delete_artifact(
    api_key: ApiKey,
    artifact_id: str,
):
    query_string = f"""
    mutation {{
        delete_artifact(
            artifact_id: "{artifact_id}",
        )
        {{
            ok
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["delete_artifact"]


@with_api_key
def update_artifact(
    api_key: ApiKey,
    artifact_id: str,
    name: Optional[str] = None,
    description: Optional[str] = None,
    s3_key: Optional[str] = None,
):
    update_str = ""
    if name:
        update_str += f'name: "{name}",\n'
    if description:
        update_str += f'description: "{description}",\n'
    if s3_key:
        update_str += f's3_key: "{s3_key}",\n'

    query_string = f"""
    mutation {{
        update_artifact(
            artifact_id: "{artifact_id}",
            {update_str}
        )
        {{
            id
            name
            description
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["update_artifact"]


@with_api_key
def artifact_url(
    api_key: ApiKey,
    artifact_id: str,
):
    query_string = f"""{{
        artifact(
            id: "{artifact_id}",
        )
        {{
            id, url
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["artifact"]["url"]


@with_api_key
def artifact_links(
    api_key: ApiKey,
    artifact_id: str,
):
    query_string = f"""{{
        artifact(
            id: "{artifact_id}",
        )
        {{
            model_ids: oracle_ids
            model_version_ids: oracle_version_ids
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["artifact"]


@with_api_key
def model_linked_artifacts(
    api_key: ApiKey,
    model_id: str,
):
    query_string = f"""{{
        model(
            id: "{model_id}",
        )
        {{
            artifacts {{
                artifact_id: id,
                name,
                description
            }}
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model"]["artifacts"]


@with_api_key
def model_version_linked_artifacts(
    api_key: ApiKey,
    model_version_id: str,
):
    query_string = f"""{{
        model_version(
            id: "{model_version_id}",
        )
        {{
            artifacts {{
                artifact_id: id,
                name,
                description
            }}
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model_version"]["artifacts"]


@with_api_key_or_jwt
def predict_for_model(
    auth_token: AuthToken,
    model_id: str,
    inputs: Union[List, Dict],
    metadata: Optional[List[Dict]] = None,
    prediction_only: bool = True,
    v2: bool = False,
) -> Union[List[List], Dict]:
    """Call the model's predict given the input json.

    Args:
        auth_token (AuthToken)
        model_id (str)
        inputs (List)
        metadata (List[Dict]): Metadata key/value pairs (e.g. name, url), one for each input.
        prediction_only(bool)

    Raises:
        RequestException: If there was an error communicating with the server.
    """

    # We have a special predict endpoint for internal use that we use
    # if a JWT is present.
    if isinstance(auth_token, ApiKey):
        predict_url = f"{settings.get_server_url()}/models/{model_id}/predict"
    elif isinstance(auth_token, JWT):
        predict_url = f"{settings.get_server_url()}/models/{model_id}/predict_internal"

    return _predict(auth_token, predict_url, inputs, metadata, prediction_only, v2=v2)


@with_api_key
def predict_for_model_version(
    api_key: ApiKey,
    model_version_id: str,
    inputs: Union[List, Dict],
    metadata: Optional[List[Dict]] = None,
    prediction_only: bool = True,
    v2: bool = False,
) -> Union[List[List], Dict]:
    """Call the model version's predict given the input json.

    Args:
        api_key (str)
        model_version_id (str)
        inputs (List)
        metadata (List[Dict]): Metadata key/value pairs (e.g. name, url), one for each input.

    Raises:
        RequestException: If there was an error communicating with the server.
    """
    predict_url = f"{settings.get_server_url()}/model_versions/{model_version_id}/predict"
    return _predict(api_key, predict_url, inputs, metadata, prediction_only, v2=v2)


@with_api_key
def set_primary(api_key, model_version_id: str):
    """Promote this version of the model as the primary version.

    Args:
        api_key (str)
        model_version_id (str)
    """
    query_string = f"""
    mutation {{
      update_model_version(model_version_id: "{model_version_id}", is_primary: true) {{
        id,
        is_primary,
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["update_model_version"]


@with_api_key
def update_model_features(
    api_key: ApiKey, model_version_id: str, feature_names: List, class_labels: Optional[List] = None
):
    """Update the feature names for the model.

    Args:
        api_key (str)
        model_version_id (str)
        feature_names (List)
        class_labels (Optional[List]): applies only to classifiers.
    """
    encoded_feature_names = base64_encoded_json_str(feature_names)
    encoded_class_labels = base64_encoded_json_str(class_labels)
    query_string = f"""
    mutation {{
      update_model_version(model_version_id: "{model_version_id}",
                           encoded_feature_names: "{encoded_feature_names}",
                           encoded_class_labels: "{encoded_class_labels}") {{
        id,
        feature_names,
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["update_model_version"]


@with_api_key
def install_requirements(api_key, requirements_txt):
    escaped_requirements_txt = requirements_txt.replace(
        "\n", "\\n"
    )  # Otherwise the mutation becomes invalid graphql.
    query_string = f"""
    mutation {{
      create_pynode_requirement(requirements_txt: "{escaped_requirements_txt}") {{
        id
        status
        error_message
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["create_pynode_requirement"]


@with_api_key
def requirement_status(api_key, requirement_id):
    query_string = f"""
    {{
      pynode_requirement(id: "{requirement_id}") {{
        id
        status
        error_message
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["pynode_requirement"]


def _format_inputs_for_v1(
    auth_token: AuthToken, inputs: Union[List, Dict], metadata: Optional[List[Dict]]
) -> Union[List, Dict]:
    """
    For internal (JWT) token use-cases, we do not need to package up the inputs.
    For external cases, for the v1 model serving path, we expect inputs to be
    provided in the format { inputs: ..., metadata: ... }.
    """
    if isinstance(auth_token, ApiKey):
        return {"inputs": inputs, "metadata": metadata}
    else:
        return inputs


def _predict(
    auth_token: AuthToken,
    predict_url: str,
    inputs: Union[List, Dict],
    metadata: Optional[List[Dict]] = None,
    prediction_only: bool = True,
    v2: bool = False,
):
    if v2:
        resp = _post_rest_query(auth_token, predict_url, inputs)
        resp_json = json.loads(resp.content)
        return resp_json["model_output"]

    resp = _post_rest_query(
        auth_token, predict_url, _format_inputs_for_v1(auth_token, inputs, metadata)
    )
    resp_json = json.loads(resp.content)
    return resp_json["predictions"] if prediction_only else resp_json


def _post_graphql_query(auth_token: AuthToken, query_string: str) -> dict:
    headers = auth_token.headers()

    resp = requests.post(
        f"{settings.get_server_url()}/graphql/", data={"query": query_string}, headers=headers
    )
    if not resp.ok:
        logger.error(f"GraphQL endpoint failed with error: {resp.content}")  # type: ignore
        resp.raise_for_status()
    resp_dict = resp.json()
    errors = resp_dict.get("errors")
    if errors:
        raise ApiError(errors[0]["message"], resp)
    return resp_dict


def _post_rest_query(auth_token: AuthToken, url: str, post_body_dict: Union[List, Dict]):
    headers = auth_token.headers()

    resp = requests.post(url, json=post_body_dict, headers=headers)
    resp.raise_for_status()
    return resp


def upload_model(serialize_file: IO, file_ext: str, file_name: str) -> str:
    """Uploads the serialized model to the appropriate environment

    Args:
        serialize_file (file): A file-like object that is the serialized representation of the model object.
        file_ext (str): The file extension for the saved model

    Returns:
        str: The key for the uploaded model

    Raises:
        RequestException: If there was an error communicating with the server.
    """

    model_file_name = f"{file_name}.{file_ext}"
    this_signed_s3_upload_post = signed_s3_upload_post(model_file_name)
    logger.debug(f"Signed s3 upload post:\n{json.dumps(this_signed_s3_upload_post, indent=4)}")

    form_fields = this_signed_s3_upload_post["form_fields"]
    form_fields["AWSAccessKeyId"] = form_fields.pop(
        "aws_access_key_id"
    )  # S3 expects key name AWSAccessKeyId
    form_fields["file"] = (file_name, serialize_file)
    logger.info("🚀 Uploading model to BaseTen 🚀")

    return _upload_file(this_signed_s3_upload_post, form_fields)


def upload_artifact(serialize_file: IO, file_ext: str, file_name: str, file_size: int) -> str:
    """Uploads the serialized artifact to the appropriate environment

    Args:
        serialize_file (file): A file-like object that is the serialized representation of the artifact object.
        file_ext (str): The file extension for the saved artifact

    Returns:
        str: The key for the uploaded artifact

    Raises:
        RequestException: If there was an error communicating with the server.
    """
    artifact_file_name = f"{file_name}.{file_ext}"
    this_signed_s3_upload_post = artifact_signed_s3_upload_post(artifact_file_name, file_size)
    logger.debug(f"Signed s3 upload post:\n{json.dumps(this_signed_s3_upload_post, indent=4)}")

    form_fields = this_signed_s3_upload_post["form_fields"]
    form_fields["AWSAccessKeyId"] = form_fields.pop(
        "aws_access_key_id"
    )  # S3 expects key name AWSAccessKeyId
    form_fields["file"] = (file_name, serialize_file)
    logger.info("🚀 Uploading artifact to BaseTen 🚀")

    return _upload_file(this_signed_s3_upload_post, form_fields)


def upload_user_file(io_stream: IO, file_name: str) -> str:
    signed_s3_upload_post = user_files_signed_s3_upload_post(file_name)
    logger.debug(f"Signed s3 upload post:\n{json.dumps(signed_s3_upload_post, indent=4)}")

    form_fields = signed_s3_upload_post["form_fields"]
    form_fields["AWSAccessKeyId"] = form_fields.pop(
        "aws_access_key_id"
    )  # S3 expects key name AWSAccessKeyId
    form_fields["file"] = (file_name, io_stream)

    _upload_file(signed_s3_upload_post, form_fields)

    return signed_s3_upload_post["get_url"]


def upload_user_dataset_file(file_path: str, file_name: str) -> dict:
    signed_s3_upload_post = user_files_signed_s3_upload_post(file_name)
    logger.debug(f"Signed s3 upload post:\n{json.dumps(signed_s3_upload_post, indent=4)}")

    form_fields = signed_s3_upload_post["form_fields"]
    form_fields["AWSAccessKeyId"] = form_fields.pop(
        "aws_access_key_id"
    )  # S3 expects key name AWSAccessKeyId
    form_fields["file"] = (file_name, open(file_path, "rb"))

    _upload_file(signed_s3_upload_post, form_fields)

    return signed_s3_upload_post


def _upload_file(this_signed_s3_upload_post: dict, form_fields: dict) -> str:
    encoder = MultipartEncoder(fields=form_fields)
    encoder_len = encoder.len
    pbar = tqdm(
        total=encoder_len,
        unit_scale=True,
        bar_format="Upload Progress: {percentage:3.0f}%% |%s{bar:100}%s| {n_fmt}/{total_fmt}"
        % (Fore.BLUE, Fore.RESET),
    )

    def callback(monitor):
        progress = monitor.bytes_read - pbar.n
        pbar.update(progress)

    monitor = MultipartEncoderMonitor(encoder, callback)
    resp = requests.post(
        this_signed_s3_upload_post["url"],
        data=monitor,
        headers={"Content-Type": monitor.content_type},
    )
    resp.raise_for_status()
    pbar.close()
    logger.info("🔮 Upload successful!🔮")

    logger.debug(f"File upload HTTP status code: {resp.status_code} and content:\n{resp.content}")  # type: ignore

    return this_signed_s3_upload_post["form_fields"]["key"]


def _tar_a_list_of_files(list_of_files: List[str]) -> tempfile._TemporaryFileWrapper:
    temp_file = tempfile.NamedTemporaryFile(suffix=".tgz")
    with tarfile.open(temp_file.name, "w:gz") as tar:
        for file in list_of_files:
            tar.add(file)
    temp_file.file.seek(0)
    return temp_file


def serialize_artifact_to_s3(name, list_of_files):
    logger.info(f"Serializing {Fore.BLUE}{name}{Style.RESET_ALL} artifact.")
    artifact_tgz = _tar_a_list_of_files(list_of_files)
    logger.info("Making contact with BaseTen 👋 👽")
    s3_key = upload_artifact(
        artifact_tgz, "tgz", ARTIFACT_FILENAME, os.path.getsize(artifact_tgz.name)
    )
    return s3_key


@with_api_key
def deactivate_model_version(api_key, model_version_id):
    query_string = f"""
    mutation {{
        deactivate_model_version(model_version_id: "{model_version_id}") {{
          ok
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["deactivate_model_version"]["ok"]


@with_api_key
def activate_model_version(api_key, model_version_id):
    query_string = f"""
    mutation {{
        activate_model_version(model_version_id: "{model_version_id}") {{
          ok
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["activate_model_version"]["ok"]


@with_api_key_or_jwt
def create_pretrained_model(auth_token: AuthToken, model_zoo_name: str, model_name: str):
    query_string = f"""
    mutation {{
        create_pretrained_model(
            model_zoo_name: "{model_zoo_name}",
            name: "{model_name}",
            description : "",
            create_app: false
        ) {{
            model {{
                created
                id
                name
                platformType: model_platform_type
                numberOfVersions: number_of_versions
                description: description
                primaryVersion: primary_version {{
                    id
                    model_id
                    modelFramework: model_framework_display_name
                    truss_spec_version
                }}
            }}
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["create_pretrained_model"]["model"]


@with_api_key_or_jwt
def create_training_dataset(
    auth_token: AuthToken, dataset_name: str, s3_key: str, training_type: str
) -> str:
    query_string = f"""
    mutation {{
        create_training_dataset(
            name: "{dataset_name}",
            s3_key: "{s3_key}",
            training_type : "{training_type}"
        ) {{
            id
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["create_training_dataset"]["id"]


@with_api_key_or_jwt
def get_training_dataset(auth_token: AuthToken, dataset_id: str) -> dict:
    query_string = f"""
    {{
        training_dataset(id: "{dataset_id}") {{
            name
            id
            presigned_s3_get_url
            training_type
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["training_dataset"]


@with_api_key_or_jwt
def get_training_run(auth_token: AuthToken, training_run_id: str) -> Dict:
    query_string = f"""
    {{
        training_run(id: "{training_run_id}") {{
            id
            status
            trained_model_name
            variables {{
                key
                value
            }}
            created
            started
            stopped
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["training_run"]


@with_api_key_or_jwt
def get_all_training_runs(auth_token: AuthToken) -> List[Dict]:
    query_string = """
    query {
        training_runs {
            id
            trained_model_name
        }
    }
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["training_runs"]


@with_api_key_or_jwt
def finetune_zoo_model(
    auth_token: AuthToken, trained_model_name: str, train_truss_name: str, encoded_variables: str
) -> str:
    query_string = f"""
    mutation {{
        finetune_model_zoo_model(
            truss_name: "{train_truss_name}",
            encoded_variables: "{encoded_variables}",
            trained_model_name: "{trained_model_name}"
        ) {{
            id
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["finetune_model_zoo_model"]["id"]


@with_api_key_or_jwt
def deploy_from_training_run(
    auth_token: AuthToken,
    training_run_id: str,
    name: str,
    idle_time_minutes: int,
) -> Tuple[str, str]:
    query_string = f"""
    mutation {{
        deploy_from_training_run(
            training_run_id: "{training_run_id}",
            name: "{name}",
            idle_time_minutes: {idle_time_minutes},
        ) {{
            version_id
            id

        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    deploy_data = resp["data"]["deploy_from_training_run"]
    return deploy_data["id"], deploy_data["version_id"]


@with_api_key_or_jwt
def cancel_training_run(
    auth_token: AuthToken,
    training_run_id: str,
) -> bool:
    query_string = f"""
    mutation {{
        cancel_training_run(
            training_run_id: "{training_run_id}",
        ) {{
            cancelled
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    deploy_data = resp["data"]["cancel_training_run"]
    return deploy_data["cancelled"]


@with_api_key_or_jwt
def get_training_logs(
    auth_token: AuthToken,
    training_run_id: str,
    start: Optional[str] = None,
    end: Optional[str] = None,
    limit: Optional[int] = None,
    direction: Optional[str] = None,
) -> List[dict]:
    query_string = f"""
    query {{
        logs(
            log_type: TRAINING,
            entity_id: "{training_run_id}",
            {f'start: "{start}",' if start is not None else ""}
            {f'end: "{end}",' if end is not None else ""}
            {f'limit: {limit},' if limit is not None else ""}
            {f'direction: "{direction}",' if direction is not None else ""}
        ) {{
            ts
            level
            msg
        }}
    }}
    """
    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["logs"]


@with_api_key_or_jwt
def get_blueprint_projects(
    auth_token: AuthToken,
) -> List[dict]:
    query_string = """
    query {
        blueprint_workflows {
            id
            name
        }
    }
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["blueprint_workflows"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['baseten',
 'baseten.client_commands',
 'baseten.common',
 'baseten.models',
 'baseten.training']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.3.0,<10.0.0',
 'click>=7.0',
 'colorama>=0.4.3',
 'coolname>=1.1.0',
 'halo>=0.0.31,<0.0.32',
 'jinja2>=2.10.3',
 'joblib>=0.12.5',
 'pytz>=2022.7.1,<2023.0.0',
 'pyyaml>=5.1',
 'requests-toolbelt>=0.9.1,<0.10.0',
 'requests>=2.22',
 'semantic-version>=2.10.0,<3.0.0',
 'single-source>=0.3.0,<0.4.0',
 'tenacity>=8.0.1,<9.0.0',
 'tqdm>=4.62.1,<5.0.0',
 'truss>=0.3.1,<0.4.0',
 'types-pillow>=9.3.0.4,<10.0.0.0',
 'types-pytz>=2022.7.1.0,<2023.0.0.0',
 'types-pyyaml>=6.0.12.2,<7.0.0.0',
 'types-requests>=2.28.11.7,<3.0.0.0',
 'types-setuptools>=65.6.0.2,<66.0.0.0']

entry_points = \
{'console_scripts': ['baseten = baseten.cli:cli_group']}

setup_kwargs = {
    'name': 'baseten',
    'version': '0.3.5rc1',
    'description': 'Deploy machine learning models to Baseten',
    'long_description': '![Baseten](https://cdn.baseten.co/docs/production/PyPiHeaderBaseten.jpeg)\n\nThis package contains the Python client and CLI for Baseten.\n\nBaseten is an ML Application Builder for data science and machine learning teams. We simplify the set-up of back-ends, front-ends, and MLOps—so your models get used faster. Baseten makes it simple to serve your machine learning models, integrate with custom business logic, and design powerful web apps for business users, all without configuring any infrastructure or writing React or JS. API endpoints, release management, and scalability all come out of the box.\n\nVisit our official documentation at [docs.baseten.co](https://docs.baseten.co).',
    'author': 'Amir Haghighat',
    'author_email': 'amir@baseten.co',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import logging
import os
import pathlib
import tempfile
import zipfile
from enum import Enum
from io import BytesIO
from pathlib import Path
from typing import IO, Optional
from uuid import uuid4

from PIL import Image

from baseten.common.api import (
    create_training_dataset,
    upload_user_dataset_file,
    upload_user_file,
)

logger = logging.getLogger(__name__)


class DatasetTrainingType(Enum):
    DREAMBOOTH = "DREAMBOOTH"
    CLASSIC_STABLE_DIFFUSION = "CLASSIC_STABLE_DIFFUSION"


DEFAULT_FILE_PATH = "png"


def upload_to_s3(io_stream: IO, file_name: str) -> str:
    """
    Uploads any stream to S3.

    Args:
        io_stream (IO): Any IO byte stream. This could be a bytes buffer, or
        an open binary file.
        file_name (str): The file_name to use when saving the file to S3

    Returns:
        str: A URL to fetch the uploaded S3 object.
    """
    upload_user_file(io_stream, file_name)
    return upload_user_file(io_stream, file_name)


def upload_pil_to_s3(image: Image.Image, file_name: Optional[str] = None) -> str:
    """
    Uploads a PIL image object to S3.

    Args:
        image (PIL.Image): A PIL image object.
        file_name (Optional[str]): The file_name to use when saving the file to S3.
            Must have an image file extension (.jpg, .png, etc.). Can be None.

    Returns:
        str: A URL to fetch the uploaded S3 object.
    """

    # If no file_name is passed, generate a random file path
    if file_name is None:
        file_name = f"{str(uuid4())}.{DEFAULT_FILE_PATH}"

    # Get the file extension from the passed in file name.
    image_format = pathlib.Path(file_name).suffix.strip(".")

    byte_buffer = BytesIO()
    image.save(byte_buffer, format=image_format)

    return upload_to_s3(byte_buffer, file_name)


def _validate_dreambooth_dataset(dir: Path) -> None:
    if not dir.exists():
        logger.warning(f"🛑 {dir} does not exist 🛑")
        return
    if not (dir / "object").exists():
        logger.warning(f"🛑 {dir / 'object'} does not exist 🛑")
        return
    if (dir / "prior_preservation").exists():
        logger.info(
            "🛎️ Please make sure to update your finetuning config to use prior_preservation loss! 🛎️"
        )
    logger.info("🛎️ Your Dreambooth dataset is valid! 🛎️")


def _validate_full_stable_diffusion_dataset(dir: Path) -> None:
    if not dir.exists():
        raise ValueError("The dataset directory does not exist.")

    image_files = [
        file
        for file in dir.glob("**/*")
        if file.suffix in [".jpg", ".jpeg", ".png", ".bmp", ".webp"]
    ]
    text_files = [file for file in dir.glob("**/*") if file.suffix == ".txt"]
    image_file_names = [file.stem for file in image_files]
    text_file_names = [file.stem for file in text_files]

    if set(image_file_names) != set(text_file_names):
        # Find the image files that don't have a corresponding text file
        # and vice versa
        missing_text_files = set(image_file_names) - set(text_file_names)
        missing_image_files = set(text_file_names) - set(image_file_names)
        if missing_text_files:
            logger.warning(
                f"🛑 Missing text files for the following image files: {missing_text_files} 🛑"
            )
            return
        if missing_image_files:
            logger.warning(
                f"🛑 Missing image files for the following text files: {missing_image_files} 🛑"
            )
            return
        logger.warning(
            "🛑 The dataset must contain a corresponding text file for each image file and vice versa. 🛑"
        )
        return

    logger.info("🛎️ Your Stable Diffusion dataset is valid! 🛎️")


def upload_dataset(name: str, dir: Path, training_type: DatasetTrainingType) -> str:
    if training_type == DatasetTrainingType.DREAMBOOTH:
        _validate_dreambooth_dataset(dir)
    elif training_type == DatasetTrainingType.CLASSIC_STABLE_DIFFUSION:
        _validate_full_stable_diffusion_dataset(dir)

    zipfile_name = f"{name}.zip"
    with tempfile.TemporaryDirectory() as tmpdir:
        zip_file = zipfile.ZipFile(os.path.join(tmpdir, zipfile_name), "w")
        for file in dir.glob("**/*"):
            # We want to ignore .DS_STORE files on Mac
            if file.name.lower() == ".ds_store":
                continue
            zip_file.write(file, arcname=f"{dir.name}/{file.relative_to(dir)}")

        zip_file.close()
        upload_params = upload_user_dataset_file(os.path.join(tmpdir, zipfile_name), zipfile_name)

        s3_key = upload_params["form_fields"]["key"]
        return create_training_dataset(name, s3_key, training_type.value)



import click

from baseten.common import api


@click.group()
def pretrained():
    """Explore pre-trained model available on Baseten."""
    pass


@pretrained.command()
@click.option("--all", is_flag=True, default=False)
def list(all):
    """Lists all pretrained models"""
    pretty_print_pretrained(api.pretrained_models()["pretrained_models"], all)


@pretrained.command()
@click.argument("pretrained_model_name")
def get(pretrained_model_name):
    """Get pretrained model, given zoo name"""
    pretty_print_pretrained(
        [api.get_pretrained_model(pretrained_model_name)["pretrained_model"]], all
    )


def pretty_print_pretrained(models, show_all=True):
    for model in models:
        available = False
        if model["s3_key"]:
            available = True
        elif not show_all:
            continue
        click.echo(f"Name: {model['pretty_name']} | Key: {model['name']} | Available: {available}")
